#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *
 *			 Queue Status Program  ( Q )
 *
 *			   Douglas P. Kingston III
 *
 *		Ballistics Research Laboratory,  March 1982
 *
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 *	14 Sep 82  DPK	Made it more that just a directory dumper.  Added
 *			code for -l, -q, and -v.
 *
 *	6 Dec 82   DPK	Added xmalloc routine called by locking code.
 *
 *	2 Jan 82   DPK	Removed -a flag.  (made it a no-op for now)
 *			Now -v is only available to the owner or sysgrp.
 *
 *	2 Feb 83   DAW  Reformatted output.
 *
 *	11 Feb 83  DAW	Modified olduid code from prreq(), was causing
 *			adjacent requests for the same person to appear with
 *			the same name & sequence number.
 *
 *	11 Aug 83  DPK	Added -a flag to print out Delayed and empty que/dev.
 */

#include <stdio.h>
#include "queue.h"
#include "devstat.h"
#include <pwd.h>
#include "dirent.h"

#ifdef DIRBLKSIZ		/* 4BSD flavor assumed */
#define	dirent	direct
#endif

extern	char	*Qcntrldir;
extern	char	*Tmpqdir;
extern	char	*Qdatadir;
extern	char	*Admprefix;
extern	char	*Qconffile;
extern	lock_t	Dae_lock;
extern	char	*index();
extern	char	*rindex();
extern	char	*strdup();
extern	char	*ctime();
extern	char	*prreq();
extern	struct	request *gethdr();
extern	char	**findque();

int	uid;
int	gid;

char	Usage[]		= "Usage:  q [-alsv] [-q qname]\n";
char	missingfmt[]	= "Missing %s argument\n";
char	dupfmt[]	= "Duplicate %s argument\n";
int	all		= 0;
int	longflag	= 0;
int	slow		= 0;
int	verbose		= 0;
char	*queue = NULL;
char	obuf[BUFSIZ];		/* for buffered STDOUT */
char	longaheader[]	=
"Dev/Que   Request    User              File           Pri Size Submitted\n";
char	longdheader[]	=
"Dev/Que   Request    User              File           Pri Size Submitted\n";

#define	PR_ACTIVE	1
#define	PR_DELAYED	2

/*
 *		M A I N ( )
 */
main (argc, argv)
int	argc;
char	**argv;
{
	DIR	*df;			/* The control directory */
	struct	dirent *dp;
	int	fuid;			/* Uid of the control file */
	register int i;
	register char *arg;
	
	fdverify();			/* Paranoid */
	if(argc == 0) exit(111);	/* Paranoid */

	/*
	 *	Q is setuid to Root so the following chdir will work
	 */
	if (chdir (Qcntrldir) < 0)
		bomb( "Can't open %s (%s)", Qcntrldir, sys_errlist[errno] );
	uid = getuid();
	gid = getgid();

	setbuf( stdout, obuf );

	for( i = 1; i < argc; i++ ) {
		if( *argv[i] != '-' )
			break;
		arg = argv[i];
		arg++;
		while( *arg ) {
			switch( *arg ) {
			case 'a':	/* All, print delayed and empty */
				all++;
				break;

			case 'l':	/* Long, single line printout */
				longflag++;
				break;

			case 'q':
				if( argv[i+1] == 0 )
					err( missingfmt, "queue" );
				else
					i++;
				if( queue != NULL )
					err( dupfmt, "queue" );
				else if( findque( argv[i] ) == NULL)
					err("Unknown queue '%s'", argv[i]);
				else
					queue = argv[i];
				break;

			case 's':	/* Slow, do not use ordered lists */
				slow++;
				break;

			case 'v':	/* Verbose, detailed printout */
				verbose++;
				break;

			default:
				err( "Unknown flag '-%c'", *arg);
				fputs( Usage, stderr );
			}
			arg++;
		}
	}

	if( lock( &Dae_lock, NOWAIT ) == 0 ) {
		fputs("The daemon does not appear to be running.\n\n", stdout);
		slow++;
		unlock( &Dae_lock );
	}

	if( longflag && verbose )
		longflag = 0;

	if( slow ) {
		if( longflag )
			fputs( longaheader, stdout );

		if(( df = opendir (".")) == NULL )
			bomb("Cannot open %s/. (%s)\n", Qcntrldir,
				sys_errlist[errno] );
		while(( dp = readdir( df )) != NULL ) {
			if (dp->d_name[0] == 'Q')
				/* We found a control file! */
				process( dp->d_name, NULL, PR_ACTIVE );
		}
		closedir( df );
	} else
		/*
		 *  Use sorted lists produced by the daemon
		 */
		pqueues();
	exit( 0 );
}

/*
 *			PQUEUES
 *
 *	Pqueues uses the configuration file to determine which
 *	queues are active and based on this, opens the appropriate
 *	sorted files and processes each request listed.
 */
pqueues()
{
	register int	filestate;
	int	argcnt;
	int	foundreqs;
	char	qdname[PATHSIZ];   /* Double duty, queue file and cntrl */
	char	linebuf[LINESIZ];
	char	**fieldp;
	register FILE	*qf;
	register FILE	*cf;
	struct qdata qdata;
	struct devstat ds;

	if(( cf = fopen (Qconffile, "r")) == NULL )
		bomb ("Cannot open configuration file (%s: %s)",
			Qconffile, sys_errlist[errno]);

	filestate = 0;

	if( longflag )
		fputs( longaheader, stdout );

	while (fgets (linebuf, LINESIZ, cf) != NULL) {

		argcnt = parse (linebuf, &fieldp);

		/*  "Blank" line or bad line, probably a comment line */
		if (argcnt <= 0)
			continue;

		/*  Lines that start with a "-" indicate end of a section */
		if (**fieldp == '-') {
			filestate++;
			if( filestate == 1 && queue == 0 )
				fputs( "Active Devices:\n", stdout );
			else if( filestate == 2 )
				printf( "\nQueued Requests:\n", *fieldp );
			continue;
		}

		if( filestate == 1 ) {
			/*
			 *  Process a DEVICE definition
			 */
			if( argcnt < 2 || queue )
				continue;
			sprintf(qdname, "%s/D_%s", Admprefix, *fieldp);
			if(( qf = fopen( qdname, "r" )) == NULL
			  || fread(&ds, sizeof ds, 1, qf) != 1
			) {
				if( qf )
					fclose( qf );
				printf("%-9.*s [error]\n", QUESIZ, *fieldp );
				continue;
			}
			fclose( qf );
			if( ds.uds.uds_status & D_DISABLED
			  || ds.dds.dds_status & D_FAILED ) {
				printf("%-9.*s", QUESIZ, *fieldp );
				if(ds.dds.dds_pid != 0)
					printf(" [NOT empty!]");
				if(ds.uds.uds_status & D_DISABLED)
					printf(" [disabled]");
				if(ds.dds.dds_status & D_FAILED)
					printf(" [failed]");
				putc('\n', stdout);
				continue;
			}
			if ( ds.dds.dds_pid == 0 ) {
				prempty( *fieldp );
				continue;
			}
			process( ds.dds.dds_rname, *fieldp, PR_ACTIVE );
		} else if( filestate == 2 ) {
			/*
			 *  Process a QUEUE definition, see if it exists
			 */
			if( argcnt == 0 )
				continue;

			if( queue && strncmp( *fieldp, queue, QUESIZ ) != SAME)
				continue;

			sprintf( qdname, "%s/Q_%s", Admprefix, *fieldp );
			if(( qf = fopen( qdname, "r" )) == NULL ) {
				err( "Cannot open queuefile for %s (%s)",
					*fieldp, sys_errlist[errno] );
				printf("\t[error]\n");
				continue;
			}
			foundreqs = 0;
			while( fread( &qdata, sizeof qdata, 1, qf )) {
				sprintf( qdname, "Q%05.5d.%d", qdata.qd_uid,
					 qdata.qd_seqno );
				process( qdname, NULL, PR_ACTIVE );
				foundreqs = 1;
			}
			if( foundreqs == 0 )
				prempty( *fieldp );
			fclose( qf );
		} else if( filestate == 3 )
			break;
	}
	fclose( cf );

	sprintf( qdname, "%s/DELAYED", Admprefix );
	if(( qf = fopen( qdname, "r" )) == NULL ) {
		err( "Cannot open queue for DELAYED requests (%s)",
			sys_errlist[errno] );
		printf("\t[error]\n");
		return;
	}
	foundreqs = 0;
	printf( "\nDelayed Requests:\n" );
	if (longflag && all)
		fputs (longdheader, stdout);
	fflush( stdout );
	while( fread( &qdata, sizeof qdata, 1, qf )) {
		sprintf( qdname, "Q%05.5d.%d", qdata.qd_uid,
			 qdata.qd_seqno );
		if (all)
			process( qdname, NULL, PR_DELAYED );
		foundreqs++;
	}
	fclose( qf );
	if( !all && foundreqs ) {
		printf( "\t%d delayed request%s\n", foundreqs,
			foundreqs > 1 ? "s" : "");
	}
	if( foundreqs == 0 )
		prempty( "" );
	return;
}

/*
 *		P R O C E S S ( )
 */
process( cname, qdname, style )
char	*cname;
char	*qdname;	/* optional name to use instead of queue name */
int	style;
{
	FILE *cf;
	char linebuf[LINESIZ];
	struct request r;

	if ((cf = fopen (cname, "r")) == NULL) {
		perror (cname);
		return;
	}

	if (gethdr (cf, &r) == (struct request *) NULL) {
		err( "Error reading request structure!");
		fclose( cf );
		return;
	}

	if( queue && strncmp( queue, r.r_que, QUESIZ ) != SAME ) {
		/*
		 *  This isn't an interesting queue
		 */
		fclose( cf );
		return;
	}

	if( verbose )
		prverbose( &r, cf, qdname );
	else if( longflag )
		prlong( &r, cf, qdname, style );
	else
		prshort( &r, qdname );
	fclose( cf );
	return;
}

prempty( name )
char	*name;		/* name of queue/device */
{
	if(all)
		printf("%-9.*s [empty]\n", QUESIZ, name );
}

prshort( rp, qdname )
register struct request *rp;
char	*qdname;	/* optional name to use instead of queue name */
{
	/***     QUESIZ     QDIRSIZ     ***/
	printf( "%-*.*s\t%-14.14s", QUESIZ, QUESIZ,
		( qdname == NULL ? rp->r_que : qdname ), prreq(rp->r_name) );
	if( rp->r_status & SHOLD )
		fputs( "  Held", stdout );
	if( rp->r_status & SKILL )
		fputs( "  Killed", stdout );
	if( rp->r_status & SDELAY )
		fputs( "  Delayed", stdout );
	putchar( '\n' );
}

/*
 *		P R L O N G
 *	Print a single line about each entry.
 */
prlong( rp, cf, qdname, style )
register struct request *rp;
FILE *cf;
char	*qdname;	/* optional name to use instead of queue name */
int	style;
{
	char	linebuf[LINESIZ];
	char	*titlep = NULL;
	char	*userp = NULL;
	long	size = 0L;
	char	*cp;

	while (fgets (linebuf, LINESIZ, cf) != NULL) {
		if (titlep == NULL && linebuf[0] == 'T') {
			register int i;

			if (cp = index( linebuf, '\n' ))
				*cp = 0;
			i = strlen (&linebuf[1]);
			if (i > 14)
				titlep = &linebuf[1] + i - 14;
			else {
				titlep = &linebuf[1];
				if (cp = rindex( titlep, '/' ))
					if (*(cp+1))
						titlep = cp+1;
			}
			titlep = strdup(titlep);
		}
		if (userp == NULL && linebuf[0] == 'U') {
			register int i;

			if (cp = index( linebuf, '\n' ))
				*cp = 0;
			userp = strdup(&linebuf[1]);
		}
		if (linebuf[0] == 'C') {
			size += (long)atol(&linebuf[1]);
		}
	}
	printf( "%-9.*s %-10.*s %-17.17s %-14.14s %3d %3dk",
		QUESIZ, ( qdname == NULL ? rp->r_que : qdname ),
		QDIRSIZ, prreq(rp->r_name),
		userp,
		titlep ? titlep : "",
		rp->r_pri,
		(int)((size+1023L)/1024) );

	if( rp->r_status & SHOLD )
		fputs( "  Held", stdout );
	if( rp->r_status & SKILL )
		fputs( "  Killed", stdout );
	if( rp->r_status & SDELAY )
		fputs( "  Delayed", stdout );
	if( (rp->r_status & (SHOLD|SKILL)) == 0 ) {
		if (style == PR_ACTIVE)
			printf (" %.12s", ctime( &rp->r_tsubmit )+4 );
		else
			printf (" %.12s", ctime( &rp->r_tstart )+4 );
	}
	fputc( '\n', stdout );
	if (titlep)
		free (titlep);
	if (userp)
		free (userp);
}

/*
 *		P R V E R B O S E ( )
 *  This function prints out the request structure
 */
prverbose( rp, cf, qdname )
struct request *rp;
FILE	*cf;
char	*qdname;	/* optional name to use instead of queue name */
{
	char	linebuf[LINESIZ];
	int	uid;

	printf("\nFile %.*s (%s)\n:", QDIRSIZ, rp->r_name, prreq(rp->r_name));
	printf ("r_name: %.*s, r_que: %.*s, r_pri: %d\n",
		QDIRSIZ, rp->r_name, QUESIZ, rp->r_que, rp->r_pri);
	printf ("r_status: %06.6o, r_uid: %d, r_gid: %d, r_forms %.*s\n",
		rp->r_status, rp->r_uid, rp->r_gid, FORMSIZ, rp->r_forms );
	printf ("r_options: %06.6o, r_copies: %d, r_size: %ld\n",
		rp->r_options, rp->r_copies, rp->r_size);
	printf ("r_submit: %s", ctime (&rp->r_tsubmit));
	printf ("r_start: %s", ctime (&rp->r_tstart));
	uid = getuid();
	if( rp->r_uid == uid || uid == 0 || mdqsmanager() )
		while (fgets (linebuf, LINESIZ, cf) != NULL)
			fputs (linebuf, stdout);

	else
		fputs( "----- Access denied to remainder of control file -----\n", stdout );
	fputc( '\n', stdout );
}

/*
 *		P R R E Q ( )
 *
 *	Pretty print a request name.
 */
char *
prreq( rname )
register char *rname;
{
	static int olduid = -1;
	static char username[20];	/* Must hold username, null */
	static char pretty[20];		/* Must hold username, ., seqno, \0 */
	char	*cp;
	int	uid;
	struct passwd *pw;
	struct passwd *getpwuid();

	if(( uid = atoi( &rname[QN_UID] )) < 0 )
		return( rname );
	if( uid != olduid )
	{
		if(( pw = getpwuid( uid )) == NULL )
			return( rname );
		else
			/* save the username */
			sprintf( username, "%s", pw->pw_name );
		olduid = uid;
	}
	sprintf( pretty, "%s.%s", username, &rname[QN_SEQ] );
	return( pretty );
}
